import logging

from openpyxl import load_workbook
import pandas as pd


from osi_dump.exporter.hypervisor.hypervisor_exporter import HypervisorExporter

from osi_dump.model.hypervisor import Hypervisor

from osi_dump import util

logger = logging.getLogger(__name__)


class ExcelHypervisorExporter(HypervisorExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_hypervisors(self, hypervisors: list[Hypervisor]):
        df = pd.DataFrame([hypervisor.model_dump() for hypervisor in hypervisors])

        logger.info(f"Exporting hypervisors for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported hypervisors for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting hypervisors for {self.sheet_name} error: {e}")



import logging

from openstack.connection import Connection

from osi_dump.exporter.role_assignment.role_assignment_exporter import (
    RoleAssignmentExporter,
)
from osi_dump.exporter.role_assignment.excel_role_assignment_exporter import (
    ExcelRoleAssignmentExporter,
)

from osi_dump.importer.role_assignment.role_assignment_importer import (
    RoleAssignmentImporter,
)
from osi_dump.importer.role_assignment.openstack_role_assignment_importer import (
    OpenStackRoleAssignmentImporter,
)


from osi_dump import util

logger = logging.getLogger(__name__)


class RoleAssignmentBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[
            tuple[RoleAssignmentImporter, RoleAssignmentExporter]
        ] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackRoleAssignmentImporter(connection)

            sheet_name = (
                f"{util.extract_hostname(connection.auth['auth_url'])}-role_ass"
            )
            exporter = ExcelRoleAssignmentExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(
        self, importer: RoleAssignmentImporter, exporter: RoleAssignmentExporter
    ):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):

        for importer, exporter in self._importer_exporter_list:
            try:

                role_assignments = importer.import_role_assignments()

                exporter.export_role_assignments(role_assignments=role_assignments)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



from typing import Optional
from pydantic import BaseModel, ConfigDict, ValidationError


class Hypervisor(BaseModel):
    hypervisor_id: str
    hypervisor_type: str
    name: str
    state: str
    status: str

    vcpus: int
    vcpus_usage: int

    memory_size: int
    memory_usage: int

    local_disk_usage: int
    local_disk_size: int

    availability_zone: Optional[str]
    aggregate_name: Optional[str]
    aggregate_id: Optional[int]

    vm_count: int



from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationError


class Volume(BaseModel):
    model_config = ConfigDict(strict=True)

    volume_id: str

    project_id: Optional[str]

    volume_name: Optional[str]

    attachments: Optional[list[str]]

    status: str

    type: str
    size: int

    snapshots: Optional[list[str]]

    updated_at: Optional[str]
    created_at: Optional[str]



import logging

from openpyxl import load_workbook

logger = logging.getLogger(__name__)


def excel_sort_sheet(file_path):
    """
    Sorts sheets in an Excel workbook alphabetically and saves the workbook.

    Args:
        file_path (str): The path to the Excel file to be sorted.
    """
    try:
        # Load the existing workbook
        workbook = load_workbook(file_path)

        # Get the current sheet names
        sheet_names = workbook.sheetnames

        # Sort sheet names alphabetically
        sorted_sheet_names = sorted(sheet_names)

        # Reorder sheets in the workbook
        for index, sheet_name in enumerate(sorted_sheet_names):
            sheet = workbook[sheet_name]
            workbook._sheets.remove(sheet)  # Remove the sheet from its current position
            workbook._sheets.insert(index, sheet)  # Insert it at the new position

        # Save the workbook with the sheets reordered
        workbook.save(file_path)

    except Exception as e:
        logger.warning(f"Error in sorting sheet: {e}")



import logging

import os

from pathlib import Path

import typer

from typing_extensions import Annotated

from osi_dump.batch_handler.external_port_batch_handler import ExternalPortBatchHandler
from osi_dump.batch_handler.load_balancer_batch_handler import LoadBalancerBatchHandler
from osi_dump.batch_handler.role_assignment_batch_handler import (
    RoleAssignmentBatchHandler,
)
from osi_dump.batch_handler.router_batch_handler import RouterBatchHandler

app = typer.Typer()


from osi_dump.batch_handler.flavor_batch_handler import FlavorBatchHandler
from osi_dump.batch_handler.image_batch_handler import ImageBatchHandler
from osi_dump.batch_handler.volume_batch_handler import VolumeBatchHandler
from osi_dump.os_connection.get_connections import get_connections


from osi_dump.batch_handler import (
    InstanceBatchHandler,
    ProjectBatchHandler,
    HypervisorBatchHandler,
    FloatingIPBatchHandler,
)


from osi_dump.importer.external_port.openstack_external_port_importer import (
    OpenStackExternalPortImporter,
)


from osi_dump import util


def _instance(connections, output_path: str):
    instance_batch_handler = InstanceBatchHandler()

    instance_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    instance_batch_handler.process()


def _floating_ip(connections, output_path: str):
    floating_ip_batch_handler = FloatingIPBatchHandler()

    floating_ip_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    floating_ip_batch_handler.process()


def _volume(connections, output_path: str):
    volume_batch_handler = VolumeBatchHandler()

    volume_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    volume_batch_handler.process()


def _project(connections, output_path: str):
    project_batch_handler = ProjectBatchHandler()

    project_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    project_batch_handler.process()


def _hypervisor(connections, output_path: str):
    hypervisor_batch_handler = HypervisorBatchHandler()

    hypervisor_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    hypervisor_batch_handler.process()


def _image(connections, output_path: str):
    image_batch_handler = ImageBatchHandler()

    image_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    image_batch_handler.process()


def _flavor(connections, output_path: str):
    flavor_batch_handler = FlavorBatchHandler()

    flavor_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    flavor_batch_handler.process()


def _role_assignment(connections, output_path: str):
    _role_assignment_batch_handler = RoleAssignmentBatchHandler()

    _role_assignment_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    _role_assignment_batch_handler.process()


def _load_balancer(connections, output_path: str):
    _load_balancer_batch_handler = LoadBalancerBatchHandler()

    _load_balancer_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    _load_balancer_batch_handler.process()


def _router(connections, output_path: str):
    _router_batch_handler = RouterBatchHandler()

    _router_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    _router_batch_handler.process()


def _external_port(connections, output_path: str):
    _external_batch_handler = ExternalPortBatchHandler()

    _external_batch_handler.add_importer_exporter_from_openstack_connections(
        connections, output_file=output_path
    )

    _external_batch_handler.process()


def inner_main(file_path: str, output_path: str):

    logger = logging.getLogger(__name__)

    connections = get_connections(file_path=file_path)

    _instance(connections=connections, output_path=output_path)

    _floating_ip(connections=connections, output_path=output_path)

    _volume(connections=connections, output_path=output_path)

    _hypervisor(connections=connections, output_path=output_path)

    _project(connections=connections, output_path=output_path)

    _image(connections=connections, output_path=output_path)

    _flavor(connections=connections, output_path=output_path)

    _role_assignment(connections=connections, output_path=output_path)

    _load_balancer(connections=connections, output_path=output_path)

    _router(connections=connections, output_path=output_path)

    _external_port(connections=connections, output_path=output_path)

    util.excel_autosize_column(output_path)

    util.excel_sort_sheet(output_path)

    logger.info(
        f"Exported OpenStack information to file: {os.path.abspath(output_path)}"
    )


def main(
    file_path: Annotated[
        Path,
        typer.Argument(
            help=(
                """
            Path of the file containing OpenStack authentication information.

            The expected JSON file format is as follows:

\b
[
    {
	"auth_url": "string",
	"project_name": "string",
	"username": "string",
	"password": "string",
	"user_domain_name": "string",
	"project_domain_name": "string"
    }
]
            """
            )
        ),
    ],
    output_path: Annotated[
        Path,
        typer.Argument(
            help="""
\b
Path of the output file, will override if file already exists
             
                """
        ),
    ] = "output.xlsx",
):
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
    )

    # Suppress pool limit warning will investigate later
    logging.getLogger("urllib3").propagate = False

    logger = logging.getLogger(__name__)

    if util.validate_dir_path(file_path=output_path) is False:
        logger.error(f"Invalid path: {output_path}, folder does not exist.")
        raise typer.Exit(1)

    inner_main(file_path=file_path, output_path=output_path)


app.command()(main)

if __name__ == "__main__":
    app()



import urllib
import urllib.parse

from openstack.connection import Connection


def get_endpoint(connection: Connection, service_type: str, interface: str) -> str:
    endpoint = connection.endpoint_for(service_type=service_type, interface=interface)

    parsed_endpoint = urllib.parse.urlparse(endpoint)

    new_hostname = urllib.parse.urlparse(connection.auth["auth_url"]).hostname

    port = parsed_endpoint.port

    ret = urllib.parse.urlunparse(
        parsed_endpoint._replace(netloc=f"{new_hostname}:{port}")._replace(
            scheme="https"
        )
    )

    return ret



import logging

import concurrent

from openstack.connection import Connection
from openstack.identity.v3.project import Project as OSProject

from osi_dump.importer.project.project_importer import ProjectImporter
from osi_dump.model.project import Project

logger = logging.getLogger(__name__)


class OpenStackProjectImporter(ProjectImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_projects(self) -> list[Project]:
        """Import projects information from Openstack

        Raises:
            Exception: Raises exception if fetching project failed

        Returns:
            list[Instance]: _description_
        """

        logger.info(f"Importing projects for {self.connection.auth['auth_url']}")

        try:
            osprojects: list[OSProject] = list(self.connection.identity.projects())
        except Exception as e:
            raise Exception(
                f"Can not fetch projects for {self.connection.auth['auth_url']}"
            ) from e

        projects: list[Project] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_project_info, project)
                for project in osprojects
            ]
            for future in concurrent.futures.as_completed(futures):
                projects.append(future.result())

        logger.info(f"Imported projects for {self.connection.auth['auth_url']}")

        return projects

    def _get_project_info(self, project: OSProject) -> Project:

        try:
            compute_quotas = self.connection.compute.get_quota_set(
                project.id, usage=True
            )
        except Exception as e:
            logger.warning(f"Get compute quotas failed for {project.id} error: {e}")

        try:
            storage_quotas = self.connection.block_storage.get_quota_set(
                project.id, usage=True
            )
        except Exception as e:
            logger.warning(f"Get storage quotas failed for {project.id} error: {e}")

        project_ret = Project(
            project_id=project.id,
            project_name=project.name,
            domain_id=project.domain_id,
            enabled=project.is_enabled,
            parent_id=project.parent_id,
            usage_instance=compute_quotas.usage["instances"],
            quota_instance=compute_quotas.instances,
            usage_ram=compute_quotas.usage["ram"],
            quota_ram=compute_quotas.ram,
            usage_vcpu=compute_quotas.usage["cores"],
            quota_vcpu=compute_quotas.cores,
            usage_volume=storage_quotas.usage["volumes"],
            quota_volume=storage_quotas.volumes,
            usage_snapshot=storage_quotas.usage["snapshots"],
            quota_snapshot=storage_quotas.snapshots,
            usage_storage=storage_quotas.usage["gigabytes"],
            quota_storage=storage_quotas.gigabytes,
        )

        return project_ret



from abc import ABC, abstractmethod

from osi_dump.model.load_balancer import LoadBalancer


class LoadBalancerImporter(ABC):
    @abstractmethod
    def import_load_balancers(self) -> list[LoadBalancer]:
        pass



import json
import logging

from pydantic import ValidationError

import typer

import openstack
from openstack.connection import Connection

from osi_dump.model.authentication_info import AuthenticationInfo


logger = logging.getLogger(__name__)

TIMEOUT_SECOND = 30


def get_connections(file_path) -> list[Connection]:
    auths = _parse_authentication_info(file_path=file_path)

    logger.info("Getting connections")

    connections = []
    auth_urls = []

    for auth in auths:
        try:
            connection = openstack.connect(
                auth_url=auth.auth_url,
                project_name=auth.project_name,
                username=auth.username,
                password=auth.password,
                project_domain_name=auth.project_domain_name,
                user_domain_name=auth.user_domain_name,
            )

            connections.append(connection)
            auth_urls.append(auth.auth_url)
        except Exception as e:
            logger.warning(f"Skipping {auth.auth_url}... error: {e}")
            pass

    logger.info("Established connection success with: ")
    for auth_url in auth_urls:
        logger.info(f"{auth_url}")

    return connections


def _parse_authentication_info(file_path: str) -> list[AuthenticationInfo]:

    with open(file_path, "r") as file:
        objects = json.load(file)
        if not isinstance(objects, list) or len(objects) == 0:
            raise ValueError(
                "The JSON file must contain a list with at least one object."
            )

        try:
            ret = [AuthenticationInfo.model_validate(obj) for obj in objects]
        except ValidationError as e:
            logger.error(e.errors())
            raise typer.Exit(1)

        return ret



import logging

import concurrent

from openstack.connection import Connection
from openstack.network.v2.floating_ip import FloatingIP as OSFloatingIP

from osi_dump.importer.floating_ip.floating_ip_importer import FloatingIPImporter
from osi_dump.model.floating_ip import FloatingIP

logger = logging.getLogger(__name__)


class OpenStackFloatingIPImporter(FloatingIPImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_floating_ips(self) -> list[FloatingIP]:
        """Import instances information from Openstack

        Raises:
            Exception: Raises exception if fetching server failed

        Returns:
            list[Instance]: _description_
        """

        logger.info(f"Importing floating ips for {self.connection.auth['auth_url']}")

        try:
            osfloating_ips: list[OSFloatingIP] = list(
                self.connection.list_floating_ips()
            )
        except Exception as e:
            raise Exception(
                f"Can not fetch floating IPs for {self.connection.auth['auth_url']}"
            ) from e

        floating_ips: list[FloatingIP] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_floating_ip_info, osfloating_ip)
                for osfloating_ip in osfloating_ips
            ]
            for future in concurrent.futures.as_completed(futures):
                floating_ips.append(future.result())

        logger.info(f"Imported floating ips for {self.connection.auth['auth_url']}")

        return floating_ips

    def _get_floating_ip_info(self, floating_ip: OSFloatingIP) -> FloatingIP:

        ret_floating_ip = FloatingIP(
            floating_ip_id=floating_ip.id,
            project_id=floating_ip.project_id,
            floating_ip_address=floating_ip.floating_ip_address,
            floating_network=floating_ip.floating_network_id,
            fixed_ip_address=floating_ip.fixed_ip_address,
            router_id=floating_ip.router_id,
            port_id=floating_ip.port_id,
            status=floating_ip.status,
            created_at=floating_ip.created_at,
            updated_at=floating_ip.updated_at,
        )

        return ret_floating_ip



import logging

from openstack.connection import Connection

from osi_dump.exporter.hypervisor.hypervisor_exporter import HypervisorExporter
from osi_dump.exporter.hypervisor.excel_hypervisor_exporter import (
    ExcelHypervisorExporter,
)

from osi_dump.importer.hypervisor.hypervisor_importer import HypervisorImporter
from osi_dump.importer.hypervisor.openstack_hypervisor_importer import (
    OpenStackHypervisorImporter,
)


from osi_dump import util

logger = logging.getLogger(__name__)


class HypervisorBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[
            tuple[HypervisorImporter, HypervisorExporter]
        ] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackHypervisorImporter(connection)

            sheet_name = (
                f"{util.extract_hostname(connection.auth['auth_url'])}-hypervisor"
            )
            exporter = ExcelHypervisorExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(
        self, importer: HypervisorImporter, exporter: HypervisorExporter
    ):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):
        for importer, exporter in self._importer_exporter_list:
            try:
                hypervisors = importer.import_hypervisors()

                exporter.export_hypervisors(hypervisors=hypervisors)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



from abc import ABC, abstractmethod

from osi_dump.model.hypervisor import Hypervisor


class HypervisorImporter(ABC):
    @abstractmethod
    def import_hypervisors(self) -> list[Hypervisor]:
        pass



import logging

from openpyxl import load_workbook
import pandas as pd


from osi_dump.exporter.floating_ip.floating_ip_exporter import FloatingIPExporter

from osi_dump.model.floating_ip import FloatingIP

from osi_dump import util

logger = logging.getLogger(__name__)


class ExcelFloatingIPExporter(FloatingIPExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_floating_ips(self, floating_ips: list[FloatingIP]):
        df = pd.DataFrame([floating_ip.model_dump() for floating_ip in floating_ips])

        logger.info(f"Exporting floating ips for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported floating ips for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting floating ips for {self.sheet_name} error: {e}")



from pydantic import BaseModel, ConfigDict


class AuthenticationInfo(BaseModel):
    model_config = ConfigDict(strict=True)

    auth_url: str
    project_name: str
    username: str
    password: str
    user_domain_name: str
    project_domain_name: str



import logging

from openstack.connection import Connection

from osi_dump.exporter.instance.excel_instance_exporter import ExcelInstanceExporter
from osi_dump.exporter.instance.instance_exporter import InstanceExporter
from osi_dump.importer.instance.instance_importer import InstanceImporter
from osi_dump.importer.instance.openstack_instance_importer import (
    OpenStackInstanceImporter,
)


from osi_dump import util

logger = logging.getLogger(__name__)


class InstanceBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[
            tuple[InstanceImporter, InstanceExporter]
        ] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            instance_importer = OpenStackInstanceImporter(connection)

            sheet_name = (
                f"{util.extract_hostname(connection.auth['auth_url'])}-instance"
            )
            instance_exporter = ExcelInstanceExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(
                importer=instance_importer, exporter=instance_exporter
            )

    def add_importer_exporter(
        self, importer: InstanceImporter, exporter: InstanceExporter
    ):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):
        for importer, exporter in self._importer_exporter_list:
            try:
                instances = importer.import_instances()

                exporter.export_instances(instances=instances)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



import pandas as pd

import logging


from osi_dump import util

from osi_dump.exporter.load_balancer.load_balancer_exporter import (
    LoadBalancerExporter,
)

from osi_dump.model.load_balancer import LoadBalancer

logger = logging.getLogger(__name__)


class ExcelLoadBalancerExporter(LoadBalancerExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_load_balancers(self, load_balancers: list[LoadBalancer]):
        df = pd.json_normalize(
            [load_balancer.model_dump() for load_balancer in load_balancers]
        )

        df = util.expand_list_column(df, "amphoraes")

        logger.info(f"Exporting load_balancers for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported load_balancers for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting load_balancers for {self.sheet_name} error: {e}")



import logging

import concurrent

from openstack.connection import Connection
from openstack.compute.v2.flavor import Flavor as OSFlavor

from osi_dump.importer.flavor.flavor_importer import FlavorImporter
from osi_dump.model.flavor import Flavor

logger = logging.getLogger(__name__)


class OpenStackFlavorImporter(FlavorImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_flavors(self) -> list[Flavor]:
        """Import flavors information from Openstack

        Raises:
            Exception: Raises exception if fetching flavor failed

        Returns:
            list[Instance]: _description_
        """

        logger.info(f"Importing flavors for {self.connection.auth['auth_url']}")

        try:
            osflavors: list[OSFlavor] = list(self.connection.list_flavors())
        except Exception as e:
            raise Exception(
                f"Can not fetch flavors for {self.connection.auth['auth_url']}"
            ) from e

        flavors: list[Flavor] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_flavor_info, osflavor)
                for osflavor in osflavors
            ]
            for future in concurrent.futures.as_completed(futures):
                flavors.append(future.result())

        logger.info(f"Imported flavors for {self.connection.auth['auth_url']}")

        return flavors

    def _get_flavor_info(self, flavor: OSFlavor) -> Flavor:

        ret_flavor = Flavor(
            flavor_id=flavor.id,
            flavor_name=flavor.name,
            ram=flavor.ram,
            vcpus=flavor.vcpus,
            disk=flavor.disk,
            swap=flavor.swap,
            public=flavor.is_public,
            properties=flavor.extra_specs,
        )

        return ret_flavor



from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationError


class FloatingIP(BaseModel):
    model_config = ConfigDict(strict=True)

    floating_ip_id: str
    project_id: Optional[str]

    floating_network: str

    floating_ip_address: str

    fixed_ip_address: Optional[str]

    router_id: Optional[str]

    port_id: Optional[str]

    status: str
    created_at: str
    updated_at: str



import pandas as pd

import logging


from osi_dump import util

from osi_dump.exporter.role_assignment.role_assignment_exporter import (
    RoleAssignmentExporter,
)

from osi_dump.model.role_assignment import RoleAssignment

logger = logging.getLogger(__name__)


class ExcelRoleAssignmentExporter(RoleAssignmentExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_role_assignments(self, role_assignments: list[RoleAssignment]):
        df = pd.json_normalize(
            [role_assignment.model_dump() for role_assignment in role_assignments]
        )

        logger.info(f"Exporting role_assignments for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported role_assignments for {self.sheet_name}")
        except Exception as e:
            logger.warning(
                f"Exporting role_assignments for {self.sheet_name} error: {e}"
            )



from abc import ABC, abstractmethod

from osi_dump.model.role_assignment import RoleAssignment


class RoleAssignmentImporter(ABC):
    @abstractmethod
    def import_role_assignments(self) -> list[RoleAssignment]:
        pass



from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationError


class Instance(BaseModel):
    model_config = ConfigDict(strict=True)

    instance_id: str
    instance_name: Optional[str]
    project_id: Optional[str]
    project_name: Optional[str]
    domain_name: Optional[str]
    private_v4_ips: Optional[list[str]]
    floating_ip: Optional[str]
    status: str
    ram: int
    vcpus: int
    vgpus: Optional[int]
    vgpu_type: Optional[str]
    hypervisor: Optional[str]

    user_id: Optional[str]

    created_at: str
    updated_at: str



import logging

from openstack.connection import Connection

from osi_dump.exporter.floating_ip.floating_ip_exporter import FloatingIPExporter
from osi_dump.exporter.floating_ip.excel_floating_ip_exporter import (
    ExcelFloatingIPExporter,
)

from osi_dump.importer.floating_ip.floating_ip_importer import FloatingIPImporter
from osi_dump.importer.floating_ip.openstack_floating_ip_importer import (
    OpenStackFloatingIPImporter,
)


from osi_dump import util

logger = logging.getLogger(__name__)


class FloatingIPBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[
            tuple[FloatingIPImporter, FloatingIPExporter]
        ] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackFloatingIPImporter(connection)

            sheet_name = (
                f"{util.extract_hostname(connection.auth['auth_url'])}-floating-ip"
            )
            exporter = ExcelFloatingIPExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(
        self, importer: FloatingIPImporter, exporter: FloatingIPExporter
    ):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):

        for importer, exporter in self._importer_exporter_list:
            try:

                floating_ips = importer.import_floating_ips()

                exporter.export_floating_ips(floating_ips=floating_ips)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



import logging

from openstack.connection import Connection

from osi_dump.exporter.project.project_exporter import (
    ProjectExporter,
)
from osi_dump.exporter.project.excel_project_exporter import ExcelProjectExporter

from osi_dump.importer.project.project_importer import ProjectImporter
from osi_dump.importer.project.openstack_project_importer import (
    OpenStackProjectImporter,
)


from osi_dump import util

logger = logging.getLogger(__name__)


class ProjectBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[tuple[ProjectImporter, ProjectExporter]] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackProjectImporter(connection)

            sheet_name = f"{util.extract_hostname(connection.auth['auth_url'])}-project"
            exporter = ExcelProjectExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(
        self, importer: ProjectImporter, exporter: ProjectExporter
    ):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):
        for importer, exporter in self._importer_exporter_list:
            try:
                projects = importer.import_projects()

                exporter.export_projects(projects=projects)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



import os


def create_file(file_path: str):
    # Check if the file already exists
    if os.path.exists(file_path):
        # Delete the old file
        os.remove(file_path)

    # Create and write to the new file
    open(file_path, "x")



import logging

from openstack.connection import Connection

from osi_dump.exporter.load_balancer.load_balancer_exporter import (
    LoadBalancerExporter,
)
from osi_dump.exporter.load_balancer.excel_load_balancer_exporter import (
    ExcelLoadBalancerExporter,
)

from osi_dump.importer.load_balancer.load_balancer_importer import (
    LoadBalancerImporter,
)
from osi_dump.importer.load_balancer.openstack_load_balancer_importer import (
    OpenStackLoadBalancerImporter,
)


from osi_dump import util

logger = logging.getLogger(__name__)


class LoadBalancerBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[
            tuple[LoadBalancerImporter, LoadBalancerExporter]
        ] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackLoadBalancerImporter(connection)

            sheet_name = f"{util.extract_hostname(connection.auth['auth_url'])}-lb"
            exporter = ExcelLoadBalancerExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(
        self, importer: LoadBalancerImporter, exporter: LoadBalancerExporter
    ):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):

        for importer, exporter in self._importer_exporter_list:
            try:

                load_balancers = importer.import_load_balancers()

                exporter.export_load_balancers(load_balancers=load_balancers)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



from .excel_autosize_column import excel_autosize_column
from .extract_hostname import extract_hostname
from .create_file import create_file
from .export_data_excel import export_data_excel
from .excel_sort_sheet import excel_sort_sheet
from .validate_dir_path import validate_dir_path
from .panda_excel import expand_list_column



import pandas as pd

import logging

from openpyxl import load_workbook

from osi_dump import util
from osi_dump.exporter.instance.instance_exporter import InstanceExporter

from osi_dump.model.instance import Instance

logger = logging.getLogger(__name__)


class ExcelInstanceExporter(InstanceExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_instances(self, instances: list[Instance]):
        df = pd.DataFrame([instance.model_dump() for instance in instances])

        logger.info(f"Exporting instances for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported instances for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting instances for {self.sheet_name} error: {e}")



import logging

import concurrent

from openstack.connection import Connection
from openstack.compute.v2.server import Server

from openstack.compute.v2.flavor import Flavor as OSFlavor

from osi_dump.importer.instance.instance_importer import InstanceImporter
from osi_dump.model.instance import Instance

logger = logging.getLogger(__name__)


class OpenStackInstanceImporter(InstanceImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_instances(self) -> list[Instance]:
        """Import instances information from Openstack

        Raises:
            Exception: Raises exception if fetching server failed

        Returns:
            list[Instance]: _description_
        """

        logger.info(f"Importing instances for {self.connection.auth['auth_url']}")

        try:
            servers: list[Server] = list(
                self.connection.compute.servers(details=True, all_projects=True)
            )
        except Exception as e:
            raise Exception(
                f"Can not fetch instances for {self.connection.auth['auth_url']}"
            ) from e

        instances: list[Instance] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_instance_info, server) for server in servers
            ]
            for future in concurrent.futures.as_completed(futures):
                instances.append(future.result())

        logger.info(f"Imported instances for {self.connection.auth['auth_url']}")

        return instances

    def _get_instance_info(self, server: Server) -> Instance:

        project_name = None
        project_id = None
        try:
            project = self.connection.identity.get_project(server.project_id)
            project_name = project.name
            project_id = project.id
        except Exception as e:
            logger.warn(
                f"Unable to obtain project name for instance: {server.name}: {e}"
            )

        domain_name = None
        try:
            domain = self.connection.identity.get_domain(project.domain_id)
            domain_name = domain.name
        except Exception as e:
            logger.warning(
                f"Unable to obtain domain name for instance {server.name}: {e}"
            )

        # Lấy thông tin IPv4 private
        private_v4_ips = []
        floating_ip = None

        try:
            for ips in server.addresses.values():
                for ip in ips:
                    if ip["OS-EXT-IPS:type"] == "fixed":
                        private_v4_ips.append(ip["addr"])
                    elif ip["OS-EXT-IPS:type"] == "floating":
                        floating_ip = ip["addr"]
        except Exception as e:
            logger.warning(
                f"Unable to obtain IP address information for instance {server.name}: {e}"
            )

        vgpus = None
        vgpu_type = None

        vgpu_metadata_property = "pci_passthrough:alias"

        try:
            flavor: OSFlavor = self.connection.get_flavor(
                name_or_id=server.flavor["id"]
            )

            vgpu_prop: str = flavor.extra_specs[vgpu_metadata_property]

            vgpu_props = vgpu_prop.split(":")

            vgpu_type = vgpu_props[0]
            vgpus = int(vgpu_props[1])

        except Exception as e:
            pass

        instance = Instance(
            instance_id=server.id,
            instance_name=server.name,
            project_id=project_id,
            project_name=project_name,
            domain_name=domain_name,
            private_v4_ips=private_v4_ips,
            floating_ip=floating_ip,
            status=server.status,
            hypervisor=server.hypervisor_hostname,
            ram=server.flavor["ram"],
            vcpus=server.flavor["vcpus"],
            created_at=server.created_at,
            updated_at=server.updated_at,
            user_id=server.user_id,
            vgpus=vgpus,
            vgpu_type=vgpu_type,
        )

        return instance



import logging

import ipaddress

import concurrent

import numpy as np

from openstack.connection import Connection
from openstack.network.v2.port import Port as OSPort
from openstack.network.v2.subnet import Subnet as OSSubnet
from openstack.network.v2.floating_ip import FloatingIP as OSFloatingIP
from openstack.network.v2.router import Router as OSRouter

from osi_dump.importer.external_port.external_port_importer import ExternalPortImporter
from osi_dump.model.external_port import ExternalPort

import osi_dump.api.neutron as osi_neutron

logger = logging.getLogger(__name__)


class OpenStackExternalPortImporter(ExternalPortImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def _get_external_ports(self) -> list[OSPort]:

        # Get all external networks
        networks = self.connection.network.networks(
            is_router_external=True,
        )

        external_ports = []

        for network in networks:
            ex_ports = list(self.connection.network.ports(network_id=network.id))

            for ex_port in ex_ports:
                external_ports.append(ex_port)

        # Get all port on those networks

        return external_ports

    def import_external_ports(self) -> list[ExternalPort]:
        """Import external_ports information from Openstack


        Raises:
            Exception: Raises exception if fetching external_port failed

        Returns:
            list[Instance]: _description_
        """

        logger.info(f"Importing external_ports for {self.connection.auth['auth_url']}")

        try:
            os_ports: list[OSPort] = self._get_external_ports()
        except Exception as e:
            raise Exception(
                f"Can not fetch external_ports for {self.connection.auth['auth_url']} {e}"
            ) from e

        external_ports: list[ExternalPort] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_external_port_info, external_port)
                for external_port in os_ports
            ]
            for future in concurrent.futures.as_completed(futures):
                external_ports.append(future.result())

        logger.info(f"Imported external_ports for {self.connection.auth['auth_url']}")

        external_ports = sorted(
            external_ports,
            key=lambda external_port: (
                external_port.network_id,
                (
                    ipaddress.ip_address(external_port.ip_address)
                    if external_port.ip_address
                    else ipaddress.ip_address("0.0.0.0")
                ),
            ),
        )

        return external_ports

    def _get_external_port_info(self, external_port: OSPort) -> ExternalPort:

        subnet_id = None
        ip_address = None
        subnet_cidr = None

        try:
            ip_address = external_port.fixed_ips[0]["ip_address"]
        except Exception as e:
            logger.warning(f"No IP address found for port {external_port.id}")

        try:
            subnet_id = external_port.fixed_ips[0]["subnet_id"]
        except Exception as e:
            logger.warning(f"No subnet ID found for {external_port.id}")

        try:
            subnet: OSSubnet = self.connection.get_subnet_by_id(subnet_id)

            subnet_cidr = subnet.cidr
        except Exception as e:
            logger.warning(f"No subnet cidr found for port {external_port.id}")

        mapped_device_project = ["network:floatingip", "network:router_gateway"]

        project_id = (
            external_port.project_id
            if external_port.device_owner not in mapped_device_project
            else self._map_project_id(
                device_owner=external_port.device_owner,
                device_id=external_port.device_id,
            )
        )

        external_port_ret = ExternalPort(
            port_id=external_port.id,
            project_id=project_id,
            network_id=external_port.network_id,
            subnet_id=subnet_id,
            subnet_cidr=subnet_cidr,
            ip_address=ip_address,
            allowed_address_pairs=external_port.allowed_address_pairs,
            device_id=external_port.device_id,
            device_owner=external_port.device_owner,
            status=external_port.status,
        )

        return external_port_ret

    def _map_project_id(self, device_owner: str, device_id: str) -> str:

        if device_owner == "network:router_gateway":
            return self._map_project_id_router(device_id=device_id)
        elif device_owner == "network:floatingip":
            return self._map_project_id_floating(device_id=device_id)

        return ""

    def _map_project_id_router(self, device_id: str) -> str:
        project_id = osi_neutron.get_router_project(
            self.connection, router_id=device_id
        )

        return project_id

    def _map_project_id_floating(self, device_id: str) -> str:
        project_id = osi_neutron.get_floating_ip_project(
            self.connection, floating_ip_id=device_id
        )

        return project_id



import pandas as pd

import logging


from osi_dump import util
from osi_dump.exporter.image.image_exporter import ImageExporter

from osi_dump.model.image import Image

logger = logging.getLogger(__name__)


class ExcelImageExporter(ImageExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_images(self, images: list[Image]):
        df = pd.json_normalize([image.model_dump() for image in images])

        logger.info(f"Exporting images for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported images for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting images for {self.sheet_name} error: {e}")



from openpyxl import load_workbook


def excel_autosize_column(file_path: str):
    """Auto size column for all sheets in excel file

    Args:
        file_path (str): Path to excel file

    Raises:
        e: Exception when loading workbook
    """
    try:
        wb = load_workbook(file_path)
    except Exception as e:
        raise e

    # Iterate through all sheets
    for sheet_name in wb.sheetnames:
        ws = wb[sheet_name]

        # Auto-size columns
        for col in ws.columns:
            max_length = 0
            column = col[0].column_letter  # Get the column name (e.g., 'A')
            for cell in col:
                try:
                    if cell.value:
                        cell_length = len(str(cell.value))
                        if cell_length > max_length:
                            max_length = cell_length
                except:
                    pass
            adjusted_width = max_length + 2  # Adjust as needed for padding
            ws.column_dimensions[column].width = adjusted_width

    # Save the modified workbook

    wb.save(file_path)



from .instance_batch_handler import InstanceBatchHandler
from .project_batch_handler import ProjectBatchHandler
from .hypervisor_batch_handler import HypervisorBatchHandler
from .floating_ip_batch_handler import FloatingIPBatchHandler
from .volume_batch_handler import VolumeBatchHandler



import logging

import concurrent

from openstack.connection import Connection
from openstack.identity.v3.role_assignment import RoleAssignment as OSRoleAssignment

from osi_dump.importer.role_assignment.role_assignment_importer import (
    RoleAssignmentImporter,
)
from osi_dump.model.role_assignment import RoleAssignment

logger = logging.getLogger(__name__)


class OpenStackRoleAssignmentImporter(RoleAssignmentImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_role_assignments(self) -> list[RoleAssignment]:
        """Import role_assignments information from Openstack

        Raises:
            Exception: Raises exception if fetching role_assignment failed

        Returns:
            list[RoleAssignment]: _description_
        """

        logger.info(
            f"Importing role_assignments for {self.connection.auth['auth_url']}"
        )

        try:
            osrole_assignments: list[OSRoleAssignment] = list(
                self.connection.identity.role_assignments()
            )
        except Exception as e:
            raise Exception(
                f"Can not fetch role_assignments for {self.connection.auth['auth_url']}"
            ) from e

        role_assignments: list[RoleAssignment] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_role_assignment_info, role_assignment)
                for role_assignment in osrole_assignments
            ]
            for future in concurrent.futures.as_completed(futures):
                role_assignments.append(future.result())

        logger.info(f"Imported role_assignments for {self.connection.auth['auth_url']}")

        return role_assignments

    def _get_role_assignment_info(
        self, role_assignment: OSRoleAssignment
    ) -> RoleAssignment:

        user_id = None
        role_id = None

        try:
            user_id = role_assignment.user["id"]
        except Exception as e:
            logger.warning(f"Can not get user id: {e}")

        try:
            role_id = role_assignment.role["id"]
        except Exception as e:
            logger.warning(f"Can not get role id: {e}")

        user_name = None
        role_name = None

        try:
            role_name = self.connection.identity.get_role(
                role_assignment.role["id"]
            ).name

        except Exception as e:
            logger.warning(f"Can not get role name: {e}")

        try:
            user_name = self.connection.identity.get_user(
                role_assignment.user["id"]
            ).name
        except Exception as e:
            logger.warning(f"Can not get user name: {e}")

        role_assignment_ret = RoleAssignment(
            user_id=user_id,
            user_name=user_name,
            role_id=role_id,
            role_name=role_name,
            scope=role_assignment.scope,
        )

        return role_assignment_ret



from abc import ABC, abstractmethod

from osi_dump.model.external_port import ExternalPort


class ExternalPortImporter(ABC):
    @abstractmethod
    def import_external_ports(self) -> list[ExternalPort]:
        pass



import pandas as pd

import logging


from osi_dump import util

from osi_dump.exporter.volume.volume_exporter import VolumeExporter

from osi_dump.model.volume import Volume

logger = logging.getLogger(__name__)


class ExcelVolumeExporter(VolumeExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_volumes(self, volumes: list[Volume]):
        df = pd.DataFrame([volume.model_dump() for volume in volumes])

        logger.info(f"Exporting volumes for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported volumes for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting volumes for {self.sheet_name} error: {e}")



import pandas as pd

import logging

from openpyxl import load_workbook

from osi_dump import util
from osi_dump.exporter.router.router_exporter import RouterExporter

from osi_dump.model.router import Router

logger = logging.getLogger(__name__)


class ExcelRouterExporter(RouterExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_routers(self, routers: list[Router]):

        df = pd.DataFrame([router.model_dump() for router in routers])

        logger.info(f"Exporting routers for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported routers for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting routers for {self.sheet_name} error: {e}")



from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationError


class LoadBalancer(BaseModel):
    model_config = ConfigDict(strict=True)

    id: str

    load_balancer_name: Optional[str]

    provisioning_status: Optional[str]

    operating_status: Optional[str]

    amphoraes: list[dict]

    project_id: Optional[str]



import logging

import concurrent

from openstack.connection import Connection

from openstack.block_storage.v3.volume import Volume as OSVolume

from osi_dump.importer.volume.volume_importer import VolumeImporter
from osi_dump.model.volume import Volume

logger = logging.getLogger(__name__)


class OpenStackVolumeImporter(VolumeImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_volumes(self) -> list[Volume]:
        """Import hypervisors information from Openstack

        Raises:
            Exception: Raises exception if fetching hypervisor failed

        Returns:
            list[Hypervisor]: _description_
        """

        logger.info(f"Importing volumes for {self.connection.auth['auth_url']}")

        try:
            osvolumes: list[OSVolume] = list(
                self.connection.block_storage.volumes(details=True, all_projects=True)
            )
        except Exception as e:
            raise Exception(
                f"Can not fetch volumes for {self.connection.auth['auth_url']}"
            ) from e

        volumes: list[Volume] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_volume_info, volume) for volume in osvolumes
            ]
            for future in concurrent.futures.as_completed(futures):
                volumes.append(future.result())

        logger.info(f"Imported volumes for {self.connection.auth['auth_url']}")

        return volumes

    def _get_volume_info(self, volume: OSVolume) -> Volume:

        snapshots = []
        try:
            snapshots = list(
                self.connection.block_storage.snapshots(
                    details=False, all_projects=True, volume_id=volume.id
                )
            )

            snapshots = [snapshot["id"] for snapshot in snapshots]

        except Exception as e:
            logger.warning(f"Fetching snapshots failed for {volume.id} error: {e}")

        ret_volume = Volume(
            volume_id=volume.id,
            volume_name=volume.name,
            project_id=volume.project_id,
            status=volume.status,
            attachments=[att["server_id"] for att in volume.attachments],
            type=volume.volume_type,
            size=volume.size,
            snapshots=snapshots,
            updated_at=volume.updated_at,
            created_at=volume.created_at,
        )

        return ret_volume



import logging

import concurrent

from openstack.connection import Connection

from openstack.load_balancer.v2.load_balancer import LoadBalancer as OSLoadBalancer

from osi_dump.importer.load_balancer.load_balancer_importer import (
    LoadBalancerImporter,
)
from osi_dump.model.load_balancer import LoadBalancer

import osi_dump.api.octavia as octavia_api

logger = logging.getLogger(__name__)


class OpenStackLoadBalancerImporter(LoadBalancerImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_load_balancers(self) -> list[LoadBalancer]:
        """Import load_balancers information from Openstack

        Raises:
            Exception: Raises exception if fetching load_balancer failed

        Returns:
            list[LoadBalancer]: _description_
        """

        logger.info(f"Importing load_balancers for {self.connection.auth['auth_url']}")

        try:
            osload_balancers: list[OSLoadBalancer] = octavia_api.get_load_balancers(
                connection=self.connection
            )
        except Exception as e:
            raise Exception(
                f"Can not fetch load_balancers for {self.connection.auth['auth_url']} {e}"
            ) from e

        load_balancers: list[LoadBalancer] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_load_balancer_info, load_balancer)
                for load_balancer in osload_balancers
            ]
            for future in concurrent.futures.as_completed(futures):
                load_balancers.append(future.result())

        logger.info(f"Imported load_balancers for {self.connection.auth['auth_url']}")

        return load_balancers

    def _get_load_balancer_info(self, load_balancer: OSLoadBalancer) -> LoadBalancer:

        amphoraes = octavia_api.get_amphoraes(
            connection=self.connection, load_balancer_id=load_balancer["id"]
        )

        for amphorae in amphoraes:
            flavor = self.connection.get_flavor_by_id(amphorae["compute_flavor"])
            amphorae["ram"] = flavor.ram
            amphorae["vcpus"] = flavor.vcpus

        load_balancer_ret = LoadBalancer(
            id=load_balancer["id"],
            load_balancer_name=load_balancer["name"],
            operating_status=load_balancer["operating_status"],
            project_id=load_balancer["project_id"],
            provisioning_status=load_balancer["provisioning_status"],
            amphoraes=amphoraes,
        )

        return load_balancer_ret



import logging

from openstack.connection import Connection
from openstack.identity.v3.service import Service
from openstack.load_balancer.v2.load_balancer import LoadBalancer

import osi_dump.util.openstack_util as os_util

logger = logging.getLogger(__name__)


def get_floating_ip_project(connection: Connection, floating_ip_id: str):
    neutron_endpoint = os_util.get_endpoint(
        connection=connection, service_type="network", interface="public"
    )

    try:
        url = f"{neutron_endpoint}/v2.0/floatingips/{floating_ip_id}?fields=project_id"
        response = connection.session.get(url)

    except Exception as e:
        print(e)

    data = response.json()

    return data["floatingip"]["project_id"]


def get_router_project(connection: Connection, router_id: str):
    neutron_endpoint = os_util.get_endpoint(
        connection=connection, service_type="network", interface="internal"
    )

    try:
        url = f"{neutron_endpoint}/v2.0/routers/{router_id}?fields=project_id"

        response = connection.session.get(url)

    except Exception as e:
        print(e)

    data = response.json()

    return data["router"]["project_id"]



import logging

from openstack.connection import Connection

from osi_dump.exporter.flavor.flavor_exporter import FlavorExporter
from osi_dump.exporter.flavor.excel_flavor_exporter import (
    ExcelFlavorExporter,
)

from osi_dump.importer.flavor.flavor_importer import FlavorImporter
from osi_dump.importer.flavor.openstack_flavor_importer import (
    OpenStackFlavorImporter,
)


from osi_dump import util

logger = logging.getLogger(__name__)


class FlavorBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[tuple[FlavorImporter, FlavorExporter]] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackFlavorImporter(connection)

            sheet_name = f"{util.extract_hostname(connection.auth['auth_url'])}-flavor"
            exporter = ExcelFlavorExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(self, importer: FlavorImporter, exporter: FlavorExporter):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):

        for importer, exporter in self._importer_exporter_list:
            try:

                flavors = importer.import_flavors()

                exporter.export_flavors(flavors=flavors)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationError


class Image(BaseModel):
    model_config = ConfigDict(strict=True)

    image_id: str

    disk_format: str
    min_disk: int
    min_ram: int
    image_name: Optional[str]
    owner: Optional[str]

    os_distro: Optional[str]
    properties: Optional[dict]

    protected: bool
    status: str
    size: Optional[int]
    virtual_size: Optional[int]
    visibility: str

    created_at: Optional[str]
    updated_at: Optional[str]



import os


def validate_dir_path(file_path: str):
    """
    Extracts the directory path from a file path and checks if it is valid.

    Args:
        file_path (str): The file path to extract the directory from.

    Returns:
        str: The directory path if it exists, otherwise an appropriate message.
    """
    # Extract the directory path
    dir_path = os.path.dirname(file_path)

    if not dir_path:
        return True
    # Check if the directory exists
    return os.path.exists(dir_path)



import logging

import concurrent

from openstack.connection import Connection
from openstack.image.v2.image import Image as OSImage

from osi_dump.importer.image.image_importer import ImageImporter
from osi_dump.model.image import Image

logger = logging.getLogger(__name__)


class OpenStackImageImporter(ImageImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_images(self) -> list[Image]:
        """Import instances information from Openstack

        Raises:
            Exception: Raises exception if fetching server failed

        Returns:
            list[Instance]: _description_
        """

        logger.info(f"Importing images for {self.connection.auth['auth_url']}")

        try:

            os_images: list[OSImage] = list(self.connection.list_images(show_all=True))
        except Exception as e:
            raise Exception(
                f"Can not fetch images for {self.connection.auth['auth_url']}"
            ) from e

        images: list[OSImage] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_image_info, image) for image in os_images
            ]
            for future in concurrent.futures.as_completed(futures):
                images.append(future.result())

        logger.info(f"Imported images for {self.connection.auth['auth_url']}")

        return images

    def _get_image_info(self, os_image: OSImage) -> Image:

        properties: dict = os_image.properties

        properties.pop("owner_specified.openstack.md5", None)
        properties.pop("owner_specified.openstack.sha256", None)
        properties.pop("owner_specified.openstack.object", None)
        properties.pop("stores", None)

        image = Image(
            image_id=os_image.id,
            disk_format=os_image.disk_format,
            min_disk=os_image.min_disk,
            min_ram=os_image.min_ram,
            image_name=os_image.name,
            owner=os_image.owner,
            properties=os_image.properties,
            protected=os_image.is_protected,
            status=os_image.status,
            os_distro=os_image.os_distro,
            size=os_image.size,
            virtual_size=os_image.virtual_size,
            visibility=os_image.visibility,
            created_at=os_image.created_at,
            updated_at=os_image.updated_at,
        )

        return image



from openstack.connection import Connection
from openstack.identity.v3.service import Service
from openstack.load_balancer.v2.load_balancer import LoadBalancer


import osi_dump.util.openstack_util as os_util


def get_load_balancers(connection: Connection) -> list[LoadBalancer]:
    octavia_endpoint = os_util.get_endpoint(
        connection=connection, service_type="load-balancer", interface="public"
    )

    url = f"{octavia_endpoint}/v2.0/lbaas/loadbalancers"

    response = connection.session.get(url)

    data = response.json()

    return data["loadbalancers"]


def get_amphoraes(connection: Connection, load_balancer_id: str) -> list[dict]:

    octavia_endpoint = os_util.get_endpoint(
        connection=connection, service_type="load-balancer", interface="public"
    )

    url = f"{octavia_endpoint}/v2/octavia/amphorae?load_balancer_id={load_balancer_id}&fields=status&fields=compute_id&fields=compute_flavor"

    response = connection.session.get(url)

    data = response.json()

    amphoraes = data["amphorae"]

    return amphoraes



from typing import Optional

from pydantic import BaseModel, ConfigDict


class Flavor(BaseModel):
    model_config = ConfigDict(strict=True)

    flavor_id: str

    flavor_name: str

    properties: Optional[dict]

    ram: int
    vcpus: int
    disk: int
    swap: Optional[int]

    public: bool



import pandas as pd

import logging

from openpyxl import load_workbook

from osi_dump import util
from osi_dump.exporter.external_port.external_port_exporter import ExternalPortExporter

from osi_dump.model.external_port import ExternalPort

logger = logging.getLogger(__name__)


class ExcelExternalPortExporter(ExternalPortExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_external_ports(self, external_ports: list[ExternalPort]):

        df = pd.DataFrame(
            [external_port.model_dump() for external_port in external_ports]
        )

        df = util.panda_excel.expand_list_column(df, "allowed_address_pairs")

        logger.info(f"Exporting external_ports for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported external_ports for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting external_ports for {self.sheet_name} error: {e}")



import logging

from openstack.connection import Connection

from osi_dump.exporter.image.image_exporter import ImageExporter
from osi_dump.exporter.image.excel_image_exporter import (
    ExcelImageExporter,
)

from osi_dump.importer.image.image_importer import ImageImporter
from osi_dump.importer.image.openstack_image_importer import (
    OpenStackImageImporter,
)


from osi_dump import util

logger = logging.getLogger(__name__)


class ImageBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[tuple[ImageImporter, ImageExporter]] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackImageImporter(connection)

            sheet_name = f"{util.extract_hostname(connection.auth['auth_url'])}-image"
            exporter = ExcelImageExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(self, importer: ImageImporter, exporter: ImageExporter):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):

        for importer, exporter in self._importer_exporter_list:
            try:

                images = importer.import_images()

                exporter.export_images(images=images)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



import logging

import os

import pandas as pd

from openpyxl import load_workbook

logger = logging.getLogger(__name__)

BOOTSTRAP_SHEET = "sheet1"


def export_data_excel(file_path: str, sheet_name: str, df: pd.DataFrame):

    if not os.path.exists(file_path):
        with pd.ExcelWriter(path=file_path, engine="openpyxl") as writer:
            empty_df = pd.DataFrame()

            empty_df.to_excel(writer, sheet_name=BOOTSTRAP_SHEET, index=False)

    try:
        with pd.ExcelWriter(
            path=file_path, engine="openpyxl", mode="a", if_sheet_exists="replace"
        ) as writer:

            df.to_excel(writer, sheet_name=sheet_name, index=False)

    except Exception as e:
        logger.warning(e)

    workbook = load_workbook(file_path)

    if BOOTSTRAP_SHEET in workbook.sheetnames:
        del workbook[BOOTSTRAP_SHEET]
        workbook.save(file_path)



import logging

from openstack.connection import Connection

from osi_dump.exporter.router.router_exporter import (
    RouterExporter,
)
from osi_dump.exporter.router.excel_router_exporter import ExcelRouterExporter

from osi_dump.importer.router.router_importer import RouterImporter
from osi_dump.importer.router.openstack_router_importer import (
    OpenStackRouterImporter,
)


from osi_dump import util

logger = logging.getLogger(__name__)


class RouterBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[tuple[RouterImporter, RouterExporter]] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackRouterImporter(connection)

            sheet_name = f"{util.extract_hostname(connection.auth['auth_url'])}-router"
            exporter = ExcelRouterExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(self, importer: RouterImporter, exporter: RouterExporter):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):
        for importer, exporter in self._importer_exporter_list:
            try:
                routers = importer.import_routers()

                exporter.export_routers(routers=routers)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



import pandas as pd

from pandas import DataFrame


def expand_list_column(df, column):
    # Find the maximum length of the list in the column
    max_len = df[column].apply(len).max()

    # Expand each dictionary in the list into columns
    expanded_df = pd.DataFrame(
        df[column]
        .apply(lambda x: [{**item} for item in x] + [{}] * (max_len - len(x)))
        .tolist(),
        index=df.index,
    )

    # Flatten the nested dictionaries into separate columns
    expanded_df = pd.json_normalize(expanded_df.to_dict(orient="records"))

    new_columns = []
    for i in range(max_len):
        for key in expanded_df.columns[i::max_len]:
            new_key = key.split(".")[-1]  # Get the actual key name
            new_columns.append(f"{column}_{i+1}.{new_key}")

    # Rename the columns to reflect the nested structure
    expanded_df.columns = new_columns

    # Drop the original column and join the expanded columns
    df = df.drop(column, axis=1).join(expanded_df)

    return df



from abc import ABC, abstractmethod

from osi_dump.model.floating_ip import FloatingIP


class FloatingIPImporter(ABC):
    @abstractmethod
    def import_floating_ips(self) -> list[FloatingIP]:
        pass



import logging

from openstack.connection import Connection

from osi_dump.exporter.external_port.external_port_exporter import ExternalPortExporter
from osi_dump.exporter.external_port.excel_external_port_exporter import (
    ExcelExternalPortExporter,
)

from osi_dump.importer.external_port.external_port_importer import ExternalPortImporter
from osi_dump.importer.external_port.openstack_external_port_importer import (
    OpenStackExternalPortImporter,
)


from osi_dump import util

logger = logging.getLogger(__name__)


class ExternalPortBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[
            tuple[ExternalPortImporter, ExternalPortExporter]
        ] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackExternalPortImporter(connection)

            sheet_name = f"{util.extract_hostname(connection.auth['auth_url'])}-ex-port"
            exporter = ExcelExternalPortExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(
        self, importer: ExternalPortImporter, exporter: ExternalPortExporter
    ):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):

        for importer, exporter in self._importer_exporter_list:
            try:

                external_ports = importer.import_external_ports()

                exporter.export_external_ports(external_ports=external_ports)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



from openstack.connection import Connection
from openstack.identity.v3.service import Service

import osi_dump.util.openstack_util as os_util


def get_usage(connection: Connection, resource_provider_id: str):

    placement_endpoint = os_util.get_endpoint(
        connection=connection, service_type="placement", interface="public"
    )

    url = f"{placement_endpoint}/resource_providers/{resource_provider_id}/usages"

    response = connection.session.get(url)

    data = response.json()

    return data["usages"]



import logging

from openstack.connection import Connection


from osi_dump.exporter.volume.excel_volume_exporter import ExcelVolumeExporter
from osi_dump.exporter.volume.volume_exporter import VolumeExporter


from osi_dump import util
from osi_dump.importer.volume.openstack_volume_importer import OpenStackVolumeImporter
from osi_dump.importer.volume.volume_importer import VolumeImporter

logger = logging.getLogger(__name__)


class VolumeBatchHandler:
    def __init__(self):
        self._importer_exporter_list: list[tuple[VolumeImporter, VolumeExporter]] = []

    def add_importer_exporter_from_openstack_connections(
        self, connections: list[Connection], output_file: str
    ):
        for connection in connections:
            importer = OpenStackVolumeImporter(connection)

            sheet_name = f"{util.extract_hostname(connection.auth['auth_url'])}-volume"
            exporter = ExcelVolumeExporter(
                sheet_name=sheet_name, output_file=output_file
            )

            self.add_importer_exporter(importer=importer, exporter=exporter)

    def add_importer_exporter(self, importer: VolumeImporter, exporter: VolumeExporter):
        self._importer_exporter_list.append((importer, exporter))

    def process(self):
        for importer, exporter in self._importer_exporter_list:
            try:
                volumes = importer.import_volumes()

                exporter.export_volumes(volumes=volumes)
            except Exception as e:
                logger.warning(e)
                logger.warning("Skipping...")



from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationError


class Router(BaseModel):
    model_config = ConfigDict(strict=True)

    project_id: Optional[str]

    router_id: str

    name: Optional[str]

    external_net_id: Optional[str]
    external_net_ip: Optional[str]

    admin_state: Optional[bool]

    status: str

    created_at: Optional[str]
    updated_at: Optional[str]



from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationError


class ExternalPort(BaseModel):
    model_config = ConfigDict(strict=True)

    port_id: str

    project_id: Optional[str]

    subnet_id: Optional[str]

    subnet_cidr: Optional[str]

    ip_address: Optional[str]

    network_id: Optional[str]

    allowed_address_pairs: Optional[list[dict]]

    device_id: Optional[str]

    device_owner: Optional[str]

    status: Optional[str]



import logging


import pandas as pd


from osi_dump.exporter.flavor.flavor_exporter import FlavorExporter

from osi_dump.model.flavor import Flavor

from osi_dump import util

logger = logging.getLogger(__name__)


class ExcelFlavorExporter(FlavorExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_flavors(self, flavors: list[Flavor]):
        df = pd.json_normalize([flavor.model_dump() for flavor in flavors])

        logger.info(f"Exporting flavors for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported flavors for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting floating ips for {self.sheet_name} error: {e}")



import logging

import concurrent

from openstack.connection import Connection
from openstack.network.v2.router import Router as OSRouter

from osi_dump.importer.router.router_importer import (
    RouterImporter,
)
from osi_dump.model.router import Router

logger = logging.getLogger(__name__)


class OpenStackRouterImporter(RouterImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_routers(self) -> list[Router]:
        """Import routers information from Openstack

        Raises:
            Exception: Raises exception if fetching router failed

        Returns:
            list[Router]: _description_
        """

        logger.info(f"Importing routers for {self.connection.auth['auth_url']}")

        try:
            osrouters: list[OSRouter] = list(self.connection.network.routers())
        except Exception as e:
            raise Exception(
                f"Can not fetch routers for {self.connection.auth['auth_url']}"
            ) from e

        routers: list[Router] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_router_info, router) for router in osrouters
            ]
            for future in concurrent.futures.as_completed(futures):
                routers.append(future.result())

        logger.info(f"Imported routers for {self.connection.auth['auth_url']}")

        return routers

    def _get_router_info(self, router: OSRouter) -> Router:
        """
                {"network_id": "49760654-71d8-4967-8fdd-5a35d3ff78ef", "external_fixed_ips": [{"subnet_id":                                   |
        |                           | "c044a5c0-4b11-4d8d-ae5e-9ff4ce6c1be6", "ip_address": "10.0.2.188"}], "enable_snat": true}
        """

        external_net_id = None

        try:
            external_net_id = router.external_gateway_info["network_id"]
        except Exception as e:
            logger.warning(f"Could not get external net id for router: {router.id}")

        external_net_ip = None

        try:
            external_net_ip = router.external_gateway_info["external_fixed_ips"][0][
                "ip_address"
            ]

        except Exception as e:
            logger.warning(f"Could not get external net ip for router {router.id}")

        router_ret = Router(
            router_id=router.id,
            name=router.name,
            external_net_id=external_net_id,
            external_net_ip=external_net_ip,
            status=router.status,
            admin_state=router.is_admin_state_up,
            project_id=router.project_id,
            created_at=router.created_at,
            updated_at=router.updated_at,
        )

        return router_ret



from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationError


class Project(BaseModel):
    model_config = ConfigDict(strict=True)

    project_id: str
    project_name: Optional[str]
    domain_id: Optional[str]
    enabled: bool
    parent_id: Optional[str]

    usage_instance: Optional[int]
    quota_instance: Optional[int]

    usage_vcpu: Optional[int]
    quota_vcpu: Optional[int]

    usage_ram: Optional[int]
    quota_ram: Optional[int]

    usage_volume: Optional[int]
    quota_volume: Optional[int]

    usage_snapshot: Optional[int]
    quota_snapshot: Optional[int]

    usage_storage: Optional[int]
    quota_storage: Optional[int]



from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationError


class RoleAssignment(BaseModel):
    model_config = ConfigDict(strict=True)

    user_id: Optional[str]

    user_name: Optional[str]

    role_id: Optional[str]

    role_name: Optional[str]

    scope: Optional[dict]



import logging

import concurrent

from openstack.connection import Connection
from openstack.compute.v2.hypervisor import Hypervisor as OSHypervisor
from openstack.compute.v2.aggregate import Aggregate as OSAggregate

from openstack.placement.v1._proxy import Proxy as PlacementProxy
from openstack.placement.v1.resource_provider_inventory import ResourceProviderInventory

from osi_dump.importer.hypervisor.hypervisor_importer import HypervisorImporter
from osi_dump.model.hypervisor import Hypervisor

from osi_dump.api.placement import get_usage

logger = logging.getLogger(__name__)


class OpenStackHypervisorImporter(HypervisorImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_hypervisors(self) -> list[Hypervisor]:
        """Import hypervisors information from Openstack

        Raises:
            Exception: Raises exception if fetching hypervisor failed

        Returns:
            list[Hypervisor]: _description_
        """
        aggregates = list(self.connection.list_aggregates())

        try:
            oshypervisors: list[OSHypervisor] = list(
                self.connection.compute.hypervisors(details=True, with_servers=True)
            )

        except Exception as e:
            raise Exception(
                f"Can not fetch hypervisor for {self.connection.auth['auth_url']}"
            ) from e

        hypervisors: list[Hypervisor] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_hypervisor_info, hypervisor, aggregates)
                for hypervisor in oshypervisors
            ]
            for future in concurrent.futures.as_completed(futures):
                hypervisors.append(future.result())

        logger.info(f"Imported hypervisors for {self.connection.auth['auth_url']}")

        return hypervisors

    def _get_hypervisor_info(
        self, hypervisor: OSHypervisor, aggregates: list[OSAggregate]
    ) -> Hypervisor:
        aggregate = self._get_aggregate(hypervisor=hypervisor)

        aggregate_id = None
        aggregate_name = None
        availability_zone = None

        if aggregate:
            aggregate_id = aggregate.id
            aggregate_name = aggregate.name
            availability_zone = aggregate.availability_zone

        placement_proxy: PlacementProxy = self.connection.placement

        rpi: ResourceProviderInventory = list(
            placement_proxy.resource_provider_inventories(
                resource_provider=hypervisor.id
            )
        )

        usage_data = get_usage(self.connection, resource_provider_id=hypervisor.id)

        vcpu = rpi[0]
        memory = rpi[1]
        disk = rpi[2]

        ret_hypervisor = Hypervisor(
            hypervisor_id=hypervisor.id,
            hypervisor_type=hypervisor.hypervisor_type,
            name=hypervisor.name,
            state=hypervisor.state,
            status=hypervisor.status,
            local_disk_size=disk["max_unit"],
            memory_size=memory["max_unit"] + memory["reserved"],
            vcpus=vcpu["max_unit"],
            vcpus_usage=usage_data["VCPU"],
            memory_usage=usage_data["MEMORY_MB"],
            local_disk_usage=usage_data["DISK_GB"],
            vm_count=len(hypervisor.servers),
            aggregate_id=aggregate_id,
            aggregate_name=aggregate_name,
            availability_zone=availability_zone,
        )

        return ret_hypervisor

    def _get_aggregate(self, hypervisor: OSHypervisor) -> OSAggregate:
        aggregates = list(self.connection.list_aggregates())

        aggregate_hypervisors: list[OSAggregate] = []

        for aggregate in aggregates:
            if hypervisor.name in aggregate.hosts:
                aggregate_hypervisors.append(aggregate)

        for aggregate_hypervisor in aggregate_hypervisors:
            if aggregate_hypervisor.availability_zone != None:
                return aggregate_hypervisor

        return aggregate_hypervisors[0]
